﻿using DBHelper;
using Microsoft.EntityFrameworkCore;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using WCS.Core;
using WCS.Entity;
using WCS.Entity.Protocol;
using WCS.Entity.Protocol.SRM;
using WCS.Service.Extensions;
using WCS.Service.Helpers;
using WCS.Service.Log;

namespace WCS.Service.Works.Station
{
    [WorkTitle(typeof(ProductHandler), "扫码入库")]
    internal class 扫码入库 : Work<StationDeviceGroup>
    {
        protected override bool SelectDevice(WCS_DEVICE dev)
        {
            return dev.CODE == "G1028";
        }

        protected override void Do(StationDeviceGroup obj)
        {
            var timer = new Stopwatch();
            timer.Start();
            obj.EX(obj =>
                {
                    obj.BcrStationIsForbid();
                    //设备组无论单卷还是双卷都必须满足的条件
                    if (obj.Items.Any(v => v.Data.VoucherNo != v.Data2.VoucherNo)) throw new WarnException($"等待任务执行--凭证号不一致");
                    if (obj.Items.Any(v => v.Data3.Status.HasFlag(StationStatus.运行状态位))) throw new WarnException($"设备运行中");

                    //成功创建的任务
                    var finishTaskList = new List<FinishTaskList<int>>();

                    //创建对应的任务
                    DB.Do(db =>
                    {
                        var devs = obj.GetBcrValid();
                        devs.Valid();
                        var infos = devs.GetWMSInTask();
                        foreach (var item in devs)
                        {
                            var dev = item.Station;
                            var next = dev.Entity.GetPath("SRM").CODE;
                            var info = infos.FirstOrDefault(v => item.FinishCode.Contains(v.ContainerCode + "}"));
                            if (db.Default.Set<WCS_TASK>().AsNoTracking().Any(v => v.BARCODE == info.ContainerCode && v.STATUS < TaskStatus.已完成 && v.TYPE == TaskType.入库))
                                throw new WarnException($"生产条码{info.ContainerCode}存在未完成任务,请检查是否为标签卡重复使用");

                            var task = new WCS_TASK();
                            task.BARCODE = info.ContainerCode;
                            task.TYPE = TaskType.入库;
                            task.STATUS = TaskStatus.执行中;
                            task.ADDRFROM = dev.Entity.CODE;
                            task.ADDRTO = info.EndPostion;
                            task.STARTTIME = DateTime.Now;
                            task.UPDATEUSER = "WCS";
                            task.UPDATETIME = DateTime.Now;
                            task.WMSTASK = int.Parse(info.WMSTaskNum);
                            task.TaskGroupKey = info.TaskGroupKey;
                            task.ADDRNEXT = next;
                            task.HEIGHT = dev.Data2.GoodsSize;
                            task.FLOOR = 1;

                            db.Default.Set<WCS_TASK>().Add(task);
                            finishTaskList.Add(new FinishTaskList<int>(task.WMSTASK, item.Station));

                            var msg = $"下达从{dev.Entity.CODE}移动至{next}的PLC指令。";
                            msg += $"[{dev.Data.Tasknum}][{dev.Data.Goodsstart}][{dev.Data.Goodsend}][{dev.Data.VoucherNo}[{dev.Data2.VoucherNo}]";
                            task.CreateStatusLog(db, msg, this.GetType());
                        }
                        //两个任务一起创建
                        db.Default.SaveChanges();
                    });

                    //检查对应的任务是否已创建成功
                    DB.Do(db =>
                    {
                        //InfoLog.INFO_CREATETASKIN($"[{obj.Entity.CODE}]--生成任务{JsonConvert.SerializeObject(finishTaskList)}");
                        foreach (var finishTask in finishTaskList)
                        {
                            var task = db.Default.Set<WCS_TASK>().FirstOrDefault(v => v.WMSTASK == finishTask.FinishCode);
                            if (task == null) continue;

                            finishTask.Station.Data.Tasknum = task.ID;
                            finishTask.Station.Data.Goodsstart = task.ADDRFROM.ToShort();
                            finishTask.Station.Data.Goodsend = task.ADDRNEXT.ToShort();
                            finishTask.Station.Data.Goodsnum = finishTaskList.Count.ToShort();
                            finishTask.Station.Data.CmdType = IstationCmdType.扫码入库;
                            finishTask.Station.Data.VoucherNo++;
                        }
                    });

                    timer.Stop();
                    InfoLog.INFO_TIMING($"{obj.Entity.CODE}--扫码入库,耗时{timer.ElapsedMilliseconds}");
                });
        }
    }

    [WorkTitle(typeof(ProductHandler), "一楼分配巷道")]
    internal class 巷道分配 : Work<StationDeviceGroup>
    {
        protected override bool SelectDevice(WCS_DEVICE dev)
        {
            return dev.CODE == "G1030";
        }

        protected override void Do(StationDeviceGroup obj)
        {
            obj.EX(obj =>
            {
                var code = obj.Entity.CODE;
                //两个设备都必须满足的条件
                if (obj.Items.Any(v => v.Data.VoucherNo != v.Data2.VoucherNo)) throw new WarnException("等待任务执行，凭证号不一致");
                if (obj.Items.Any(v => v.Data3.Status.HasFlag(StationStatus.运行状态位))) throw new DoException("运行中");

                //成功分配巷道的任务的任务
                var finishTaskList = new List<FinishTaskList<int>>();

                //变更数据库信息
                DB.Do(db =>
                {
                    var devs = obj.Items.Where(v => v.Data2.Status.HasFlag(IstationStatus.光电状态) && v.Data2.Request == IstationRequest.请求分配目标地址)
                                           .Where(v => v.Data2.Tasknum > 10000 && v.Data2.Goodsend != 0);
                    if (!devs.Any()) return; //无可执行任务
                    var taskIds = devs.Select(dev => dev.Data2.Tasknum).ToList();
                    var taskList = db.Default.Set<WCS_TASK>().Where(v => taskIds.Contains(v.ID)).ToList();

                    taskList.ValidTaskCheck(devs.Count(), db);

                    var res = WMS.GetTunnelList(taskList.Select(v => v.WMSTASK.ToString()).ToList(), code);
                    if (string.IsNullOrEmpty(res.TunnelNum)) throw new WarnException($"WMS未返回巷道");
                    var tunnelNo = res.TunnelNum.Split(',').Select(v => "TY" + v).ToList();

                    var tunnels = Device.Where(v => tunnelNo.Contains(v.CODE)).ToList();

                    List<TunnelInfo> tunnelInfos = new List<TunnelInfo>();
                    foreach (var item in tunnels)
                    {
                        //当前巷道的取货点
                        var allIn = Device.Where(v => v.Is(DF.SRM二级品取货)) //二楼所有取货点
                                          .Where(v => v.ROUTES.Any(p => p.NEXT == item)) //下一个点为当前巷道的取货点
                                          .Select(v => v.CODE)
                                          .ToList();

                        //下一个目标地址包含取货点的设备,此时获得的数据是旋转台
                        var turntable = Device.Where(v => v.IsConv() && v.ROUTES.Any(p => p.NEXT != null && allIn.Contains(p.NEXT.CODE))).Select(v => v.CODE);

                        var putStation = Device.Where(v => v.IsConv() && v.ROUTES.Any(p => p.NEXT != null && turntable.Contains(p.NEXT.CODE))) //下一个目标地址包含取货点的设备
                                      .Select(v => v.Create<StationDevice>()) //取所有可以到达取货点设备的信息
                                      .Where(v => !v.Data3.Status.HasFlag(StationStatus.运行状态位) && !v.Data2.Status.HasFlag(IstationStatus.光电状态) && v.Data2.Tasknum < 10000)//筛选出空闲的路径点，此处因输送机都是一个动力，因此可以先找路径点再找设备组
                                      .Distinct()
                                      .FirstOrDefault();//去一次重
                        if (putStation == null) continue;
                        var taskInStation = Device.Where(p => p.CODE.StartsWith("G") && p.DEVICEGROUP.Any(d => d.MEMBER == putStation.Entity)).OrderBy(p => p.CODE).FirstOrDefault();
                        //RGV是否有正在前往这个地址的任务
                        var rgv8 = Device.Find("RGV8").Create<RGVDevice>();
                        if ((!rgv8.Data2.SystemStatus.HasFlag(WCS.Entity.Protocol.RGV.RGVRunStatus.空闲) || !rgv8.Data2.WorkMode.HasFlag(WCS.Entity.Protocol.RGV.RGVMode.自动))
                           && rgv8.Data2.DestPosition_1 == taskInStation.CODE.Replace("G", "").ToShort()) continue;
                        tunnelInfos.Add(new TunnelInfo
                        {
                            Tunnel = item,
                            taskIN = taskInStation, //找到放货点设备所在组
                            SRM = Device.Where(p => p.IsSC()).FirstOrDefault(p => item.ROUTES.Any(d => d.NEXT.CODE == p.CODE)).Create<SRMDevice>()
                        });
                    }
                    //筛选出优先级最高的可用巷道
                    var tunnelInfo = tunnelInfos.Where(v => { try { return v.SRM.Data3.SCAlarm == 0 && v.SRM.Data2.SRMMode == SCMode.远程 && v.SRM.Data2.SRMStatus == SCRunStatus.空闲; } catch { return false; } })
                                                .Where(v => !db.Default.Set<WCS_TASK>().Any(p => p.Priority > 0 && p.DEVICE == v.SRM.Entity.CODE && p.STATUS < TaskStatus.堆垛机完成))
                                                .OrderBy(v => tunnelNo.IndexOf(v.Tunnel.CODE)).FirstOrDefault();
                    if (tunnelInfo == null) throw new WarnException("无可用巷道");

                    //开始变更任务信息
                    foreach (var item in devs)
                    {
                        var dev = Device.Find(item.Entity.CODE).Create<StationDevice>();
                        var task = taskList.FirstOrDefault(p => p.ID == dev.Data2.Tasknum);
                        if (task == null) throw new WarnException($"WCS无该任务{dev.Data2.Tasknum}--{dev.Entity.CODE}");

                        task.DEVICE = tunnelInfo.SRM.Entity.CODE;
                        task.TUNNEL = tunnelInfo.Tunnel.CODE;
                        task.ADDRNEXT = tunnelInfo.taskIN.CODE;
                        task.TaskGroupKey = res.WMSTaskGroupKey;
                        task.ADDRTO = task.DEVICE;
                        var msg = $"下达从G1030移动至{dev.Data.Goodsend}的PLC指令。同时将任务分配至[{task.TUNNEL}]-[{task.DEVICE}]";
                        msg += $"[{dev.Data.Tasknum}][G1030][{dev.Data.Goodsend}][{tunnelInfo.SRM.Entity.CODE}][{dev.Data.VoucherNo}[{dev.Data2.VoucherNo}]";
                        task.CreateStatusLog(db, msg, this.GetType());
                        finishTaskList.Add(new FinishTaskList<int>(task.ID, item.Entity.Create<StationDevice>()));
                    }

                    db.Default.SaveChanges();
                });

                // 开始将任务信息写入到设备
                DB.Do(db =>
                {
                    foreach (var finishTask in finishTaskList)
                    {
                        var task = db.Default.Set<WCS_TASK>().Find(finishTask.FinishCode);
                        if (task == null) continue;

                        finishTask.Station.Data.Tasknum = task.ID;
                        finishTask.Station.Data.Goodsstart = 1030;
                        finishTask.Station.Data.Goodsend = task.ADDRNEXT.Replace("G", "").ToShort();
                        finishTask.Station.Data.Goodsnum = finishTaskList.Count.ToShort();
                        finishTask.Station.Data.CmdType = IstationCmdType.分配目标地址;
                        finishTask.Station.Data.VoucherNo++;
                    }
                });
            });
        }
    }

    [WorkTitle(typeof(ProductHandler), "一楼入库旋转台二次分配巷道")]
    internal class 一楼入库旋转台二次分配巷道 : Work<StationDeviceGroup>
    {
        protected override void Do(StationDeviceGroup obj)
        {
            obj.EX(obj =>
            {
                var code = obj.Entity.CODE;

                //两个设备都必须满足的条件
                if (obj.Items.Any(v => v.Data.VoucherNo != v.Data2.VoucherNo)) throw new WarnException("等待执行任务--凭证号不一致");
                if (obj.Items.Any(v => v.Data3.Status.HasFlag(StationStatus.运行状态位))) throw new DoException("运行中");
                //成功分配巷道的任务的任务
                var finishTaskList = new List<FinishTaskList<int>>();

                //变更数据库信息
                DB.Do(db =>
                {
                    var devs = obj.Items.Where(v => v.Data2.Status.HasFlag(IstationStatus.光电状态) && v.Data2.Request == IstationRequest.请求分配目标地址)
                                       .Where(v => v.Data2.Tasknum > 10000);
                    if (!devs.Any()) return; //无可执行任务
                    var taskIds = devs.Select(dev => dev.Data2.Tasknum).ToList();
                    var taskList = db.Default.Set<WCS_TASK>().Where(v => taskIds.Contains(v.ID)).ToList();

                    taskList.ValidTaskCheck(devs.Count(), db);

                    var res = WMS.GetTunnelList(taskList.Select(v => v.WMSTASK.ToString()).ToList(), code);
                    if (string.IsNullOrEmpty(res.TunnelNum)) throw new WarnException($"WMS未返回巷道");
                    var tunnelNo = res.TunnelNum.Split(',').Select(v => "TY" + v).ToList();

                    var tunnels = Device.Where(v => tunnelNo.Contains(v.CODE)).ToList();

                    List<TunnelInfo> tunnelInfos = new List<TunnelInfo>();
                    foreach (var item in tunnels)
                    {
                        //当前巷道的取货点
                        var allIn = Device.Where(v => v.Is(DF.SRM二级品取货)) //一楼所有取货点
                                      .Where(v => v.ROUTES.Any(p => p.NEXT == item)) //下一个点为当前巷道的取货点
                                      .Select(v => v.CODE)
                                      .ToList();
                        //下一个目标地址包含取货点的设备,此时获得的数据是旋转台
                        var turntable = Device.Where(v => v.IsConv() && v.ROUTES.Any(p => p.NEXT != null && allIn.Contains(p.NEXT.CODE)))
                                          .Where(v => obj.Items.Any(p => p.Entity == v))// 筛选出包含在当前设备组的设备
                                          .Distinct()
                                          .FirstOrDefault();//去一次重
                        if (turntable == null) continue;
                        var dev = Device.Where(p => p.CODE.StartsWith("G") && p.DEVICEGROUP.Any(d => d.MEMBER == turntable)).OrderBy(p => p.CODE).FirstOrDefault();

                        tunnelInfos.Add(new TunnelInfo
                        {
                            Tunnel = item,
                            taskIN = dev, //找到放货点设备所在组
                            SRM = Device.Where(p => p.IsSC()).FirstOrDefault(p => item.ROUTES.Any(d => d.NEXT.CODE == p.CODE)).Create<SRMDevice>()
                        });
                    }
                    //筛选出优先级最高的可用巷道
                    var tunnelInfo = tunnelInfos.Where(v => { try { return v.SRM.Data3.SCAlarm == 0 && v.SRM.Data2.SRMMode == SCMode.远程 && v.SRM.Data2.SRMStatus == SCRunStatus.空闲; } catch { return false; } })
                                            .OrderBy(v => tunnelNo.IndexOf(v.Tunnel.CODE)).FirstOrDefault();
                    if (tunnelInfo == null) throw new WarnException("无可用巷道");

                    //开始变更任务信息
                    foreach (var item in devs)
                    {
                        var dev = item.Entity.Create<StationDevice>();
                        var task = taskList.FirstOrDefault(p => p.ID == dev.Data2.Tasknum);
                        if (task == null) throw new WarnException($"WCS无该任务{dev.Data2.Tasknum}--{dev.Entity.CODE}");

                        task.DEVICE = tunnelInfo.SRM.Entity.CODE;
                        task.TUNNEL = tunnelInfo.Tunnel.CODE;
                        task.ADDRNEXT = tunnelInfo.taskIN.CODE;
                        task.TaskGroupKey = res.WMSTaskGroupKey;
                        task.ADDRTO = task.DEVICE;
                        var msg = $"下达从{obj.Entity.CODE}移动至{dev.Data.Goodsend}的PLC指令。同时将任务分配至[{task.TUNNEL}]-[{task.DEVICE}]";
                        msg += $"[{dev.Data.Tasknum}][{obj.Entity.CODE}][{dev.Data.Goodsend}][{tunnelInfo.SRM.Entity.CODE}][{dev.Data.VoucherNo}[{dev.Data2.VoucherNo}]";
                        task.CreateStatusLog(db, msg, this.GetType());
                        finishTaskList.Add(new FinishTaskList<int>(task.ID, dev));
                    }
                    db.Default.SaveChanges();
                });

                // 开始将任务信息写入到设备
                DB.Do(db =>
                {
                    foreach (var finishTask in finishTaskList)
                    {
                        var task = db.Default.Set<WCS_TASK>().Find(finishTask.FinishCode);
                        if (task == null) continue;

                        finishTask.Station.Data.Tasknum = task.ID;
                        finishTask.Station.Data.Goodsstart = obj.Entity.CODE.Replace("G", "").ToShort();
                        finishTask.Station.Data.Goodsend = finishTask.Station.Entity.GetPath(task.DEVICE).CODE.ToShort();
                        finishTask.Station.Data.Goodsnum = finishTaskList.Count.ToShort();
                        finishTask.Station.Data.CmdType = IstationCmdType.分配目标地址;
                        finishTask.Station.Data.VoucherNo++;
                    }
                });
            });
        }

        protected override bool SelectDevice(WCS_DEVICE dev)
        {
            return devCodes.Contains(dev.CODE);
        }

        private List<string> devCodes = new List<string>() { "G1038", "G1047", "G1056", "G1065" };
    }

    //[WorkTitle(typeof(ProductHandler), "一楼RGV放货结束分配目标地址")]
    internal class 一楼RGV放货结束分配目标地址 : Work<StationDeviceGroup>
    {
        protected override void Do(StationDeviceGroup obj)
        {
            obj.EX(obj =>
            {
                if (obj.Items.Any(v => v.Data.VoucherNo != v.Data2.VoucherNo)) throw new WarnException($"等待执行任务,凭证号不一致");
                if (obj.Items.Any(v => v.Data3.Status.HasFlag(StationStatus.运行状态位))) throw new DoException("运行中");
                var finishTaskList = new List<FinishTaskList<int>>();
                var devs = new List<FinishTaskList<string>>();
                foreach (var dev in obj.Items)
                {
                    if (dev.Data2.Request != IstationRequest.请求分配目标地址)
                    {
                        InfoLog.INFO_INFO($"{obj.Entity.CODE}无请求-分配目标地址--2");
                        continue;
                    }
                    if (!dev.Data2.Status.HasFlag(IstationStatus.光电状态))
                    {
                        InfoLog.INFO_WarnDb($"{dev.Entity.CODE}有请求无光电", obj.Entity.CODE, WCS_EXCEPTIONTYPE.设备异常);
                        continue;
                    }
                    devs.Add(new FinishTaskList<string>(dev.Entity.CODE, dev.Entity.Create<StationDevice>()));
                }

                DB.Do(db =>
                {
                    foreach (var dev in devs)
                    {
                        var task = db.Default.Set<WCS_TASK>().FirstOrDefault(v => v.ID == dev.Station.Data2.Tasknum);
                        var next = dev.Station.Entity.GetPath(task.ADDRTO);
                        task.ADDRNEXT = next.CODE;

                        var msg = $"下达从{dev.Station.Data.Goodsstart}移动至{dev.Station.Data.Goodsend}的PLC指令";
                        msg += $"[{dev.Station.Data.Tasknum}][{dev.Station.Data.Goodsstart}][{dev.Station.Data.Goodsend}][{dev.Station.Data.VoucherNo}[{dev.Station.Data2.VoucherNo}]";
                        task.CreateStatusLog(db, msg, this.GetType());
                        finishTaskList.Add(new FinishTaskList<int>(task.ID, dev.Station));
                    }
                    db.Default.SaveChanges();
                });

                DB.Do(db =>
                {
                    foreach (var finish in finishTaskList)
                    {
                        var task = db.Default.Set<WCS_TASK>().FirstOrDefault(v => v.ID == finish.FinishCode);
                        if (task == null) continue;
                        finish.Station.Data.Tasknum = task.ID;
                        finish.Station.Data.Goodsstart = finish.Station.Entity.CODE.ToShort();
                        finish.Station.Data.Goodsend = task.ADDRNEXT.ToShort();
                        finish.Station.Data.CmdType = IstationCmdType.分配目标地址;
                        finish.Station.Data.VoucherNo++;
                    }
                });
            });
        }

        protected override bool SelectDevice(WCS_DEVICE dev)
        {
            return devCodes.Contains(dev.CODE);
        }

        private List<string> devCodes = new List<string>() { "G1035", "G1044", "G1053", "G1062" };
    }
}